import gobject
import gtk
import numpy
import traceback
from itertools import izip

import qubx.accept
import qubx.global_namespace
import qubx.GTK
import qubx.task
import qubx.util_types
from qubx.GTK import pack_item, pack_label, pack_check, pack_scrolled, pack_space
from qubx.settings import Property, Propertied
import qubxidlmdl


@Propertied(Property('Lmin', 3, 'Width of the shortest dwell, in samples'),
            Property('split3', False, 'True to also try splitting an interval three ways before giving up'),
            Property('combine_close', True, 'True to combine adjacent events with similar Amp'),
            Property('combine_delta', 0.5, 'If combine_close, adjacent events nearer than this are merged'),
            Property('make_list', True, 'whether to output a List of each event'),
            Property('list_name', "MDL events", "name of output List"))
class MDLIdealizer(gtk.HBox):
    __explore_featured = ['global_name', 'robot', 'txtStatus', 'outStatus', 'shared_controls', 'idealize', 'ctx', 'out_list']
    def __init__(self):
        gtk.HBox.__init__(self)
        self.global_name = 'QubX.Modeling.Idealize.methods["MDL"]'
        self.__ref = qubx.util_types.Reffer()
        self.propertied_connect_settings('Idealize_MDL')
        self.robot = qubx.task.Robot('Idl:MDL', self.__ref(lambda: qubx.task.Tasks.add_task(self.robot)),
                                     self.__ref(lambda: qubx.task.Tasks.remove_task(self.robot)))
        self.robot.OnException += self.__ref(self.__onException)
        self.robot.OnInterrupt += self.__ref(self.__onInterrupt)
        
        self.set_size_request(200, 50)
        self.txtStatus = gtk.TextView()
        self.txtStatus.set_wrap_mode(gtk.WRAP_WORD)
        self.outStatus = qubx.GTK.TextViewAppender(self.txtStatus)
        self.txtStatus.set_editable(False)
        qubx.GTK.SetFixedWidth(self.txtStatus)
        pack_scrolled(self.txtStatus, self, expand=True)
        self.outStatus.write("""Unsupervised Idealization of Ion Channel Recordings by Minimum Description Length: Application to Human PIEZO1-Channels

Radhakrishnan Gnanasambandam, Morten S. Nielsen, Christopher Nicolai, Frederick Sachs, Johannes P. Hofgaard and Jakob K. Dreyer

https://doi.org/10.3389/fninf.2017.00031

Abstract:
Researchers can investigate the mechanistic and molecular basis of many physiological phenomena in cells by analyzing the fundamental properties of single ion channels. These analyses entail recording single channel currents and measuring current amplitudes and transition rates between conductance states. Since most electrophysiological recordings contain noise, the data analysis can proceed by idealizing the recordings to isolate the true currents from the noise. This de-noising can be accomplished with threshold crossing algorithms and Hidden Markov Models, but such procedures generally depend on inputs and supervision by the user, thus requiring some prior knowledge of underlying processes. Channels with unknown gating and/or functional sub-states and the presence in the recording of currents from uncorrelated background channels present substantial challenges to such analyses. Here we describe and characterize an idealization algorithm based on Rissanen's Minimum Description Length (MDL) Principle. This method uses minimal assumptions and idealizes ion channel recordings without requiring a detailed user input or a priori assumptions about channel conductance and kinetics. Furthermore, we demonstrate that correlation analysis of conductance steps can resolve properties of single ion channels in recordings contaminated by signals from multiple channels. We first validated our methods on simulated data defined with a range of different signal-to-noise levels, and then showed that our algorithm can recover channel currents and their substates from recordings with multiple channels, even under conditions of high noise. We then tested the MDL algorithm on real experimental data from human PIEZO1 channels and found that our method revealed the presence of substates with alternate conductances.
""")

        column = pack_item(gtk.VBox(), self)
        h = pack_item(gtk.HBox(), column)
        pack_label('Lmin:', h)
        txt = pack_item(qubx.GTK.NumEntry(self.Lmin, qubx.accept.acceptIntGreaterThan(0), width_chars=6), h, at_end=True)
        self.propertied_connect_NumEntry('Lmin', txt)
        h.set_tooltip_text("Width of shortest dwells, in number of samples")
        h = pack_item(gtk.HBox(), column)
        chk = pack_check('Try 3-way splits (slow)', h)
        self.propertied_connect_check('split3', chk)
        h = pack_item(gtk.HBox(), column)
        chk = pack_check('Merge adjacent events', h)
        self.propertied_connect_check('combine_close', chk)
        h = pack_item(gtk.HBox(), column)
        pack_space(h, x=20)
        pack_label('if amps closer than:', h)
        txt = pack_item(qubx.GTK.NumEntry(self.combine_delta, qubx.accept.acceptFloatGreaterThan(0.0), '%.3g'), h)
        self.propertied_connect_NumEntry('combine_delta', txt)
        h = pack_item(gtk.HBox(), column)
        chk = pack_check('make List of events', h)
        self.propertied_connect_check('make_list', chk)
        h = pack_item(gtk.HBox(), column)
        pack_label("List name:", h)
        txt = pack_item(qubx.GTK.NumEntry(self.list_name, qubx.accept.acceptStringNonempty), h, expand=True)
        self.propertied_connect_NumEntry('list_name', txt)
        self.shared_controls = pack_item(gtk.VBox(), column, at_end=True)
        
        self.ctx = qubxidlmdl.MDLIdealizeContext()
        self.out_list = None
        self.idealizer = None

    def qubx_finalize(self):
        self.robot.stop()
    def idealize(self, segments, model, on_finish):
        QubX = qubx.global_namespace.QubX
        segments = QubX.DataSource.get_segmentation()
        if segments:
            if self.make_list:
                self.out_list = QubX.Data.file.lists.show_list(self.list_name)
                self.out_list.clear()
            else:
                self.out_list = None
            self.ideal_updater = segments[0].file.update_idl(segments[0].signal) if segments else None
            self.robot.do(self.robot_idealize, segments, self.Lmin, self.split3, self.combine_close, self.combine_delta, on_finish)
        else:
            gobject.idle_add(on_finish)
    def robot_idealize(self, segments, Lmin, split3, combine_close, combine_delta, on_finish):
        idlseg = [ [] ]
        seg_amps, seg_stds = [], []
        last_seg_index = [-1]
        c = [0]
        def emit_idl():
            if len(seg_amps) > 1:
                seg_amps[:] = [numpy.hstack(seg_amps)]
                seg_stds[:] = [numpy.hstack(seg_stds)]
            if len(seg_amps):
                gobject.idle_add(self.__set_idl, segments[0].file, segments[0].signal, last_seg_index[0], seg_amps[0], seg_stds[0], idlseg[0])
                idlseg[:] = [ [] ]
                seg_amps[:] = []
                seg_stds[:] = []
        def add_idl(seg_index, amps, stds, ff, ll):
            if seg_index != last_seg_index[0]:
                emit_idl()
                last_seg_index[0] = seg_index
                c[0] = 0
            n = len(ff)
            cc = numpy.arange(c[0], c[0]+n, dtype="int32")
            c[0] += n
            idlseg[0].append((ff, ll, cc))
            seg_amps.append(amps)
            seg_stds.append(stds)
        try:
            if segments:
                self.idealizer = self.ctx.new_idealizer()
                for seg in segments:
                    for chunk in seg.chunks:
                        if chunk.included:
                            samples = chunk.get_samples().samples
                            ff, ll, amps, stds = self.idealizer.idealize(samples, chunk.f, Lmin, split3, combine_close, combine_delta)
                            #stds = numpy.array([samples[f-chunk.f:l-chunk.f-1].std() for f,l in izip(ff, ll)], dtype="float32")
                            add_idl(seg.index, amps, stds, ff, ll)
                emit_idl()
        finally:
            self.idealizer = None
            if segments:
                gobject.idle_add(self.ideal_updater.done)
            gobject.idle_add(on_finish)
    def __set_idl(self, datafile, out_ix, seg_index, amps, stds, idlseg):
        for ff, ll, cc in idlseg:
            self.ideal_updater.set_dwells(len(ff), ff, ll, cc, amps, stds, event=True)
        if not (self.out_list is None):
            lst = self.out_list
            sampling_ms = datafile.sampling * 1e3
            for ff, ll, cc in idlseg:
                last_amp = amps[cc[0]] if len(cc) else 0.0
                for f, l, c in izip(ff, ll, cc):
                    lst.insert_selection(f, l, Duration=sampling_ms*(l-f+1), Amp=amps[c], Std=stds[c], Delta=amps[c]-last_amp)
                    last_amp = amps[c]
        #buf = cStringIO.StringIO()
        #buf.write("Levels:\n")
        #for a in amps:
        #    buf.write("\t%.5g\n" % a)
        #buf.write("\n")
        #self.outStatus.write(buf.getvalue())
    def __onException(self, robot, typ, val, trace):
        traceback.print_exception(typ, val, trace, file=self.outStatus)
    def __onInterrupt(self, robot, cancel):
        if self.idealizer:
            self.idealizer.interrupt()
        cancel()

# progress
# opencl

# warn slow on many data
